#!/usr/bin/env python

import torch
import argparse
from src.loadopts import *

import numpy as np



METHOD = "WhiteBox"
FMT = "{description}={bn_adv}={attack}-{stepsize}-{steps}" \
        "={epsilon_min}-{epsilon_max}-{epsilon_times}"


parser = argparse.ArgumentParser()
parser.add_argument("model", type=str)
parser.add_argument("dataset", type=str)
parser.add_argument("info_path", type=str)
parser.add_argument("--bn_adv", action="store_false", default=True)
parser.add_argument("--attack", type=str, default="pgd-linf")
parser.add_argument("--epsilon_min", type=float, default=8/255)
parser.add_argument("--epsilon_max", type=float, default=1.)
parser.add_argument("--epsilon_times", type=int, default=1)
parser.add_argument("--stepsize", type=float, default=0.1,
                    help="pgd:rel_stepsize, cwl2:step_size, deepfool:overshoot, bb:lr")
parser.add_argument("--steps", type=int, default=20)
parser.add_argument("-b", "--batch_size", type=int, default=128)
parser.add_argument("--progress", action="store_false", default=True,
                help="show the progress if true")
parser.add_argument("--seed", type=int, default=1)
parser.add_argument("-m", "--description", type=str, default="attack")

parser.add_argument('--scatter', action='store_true',
    help='if set, scatter the adversarial sample in feature space')
parser.add_argument('--scatter_eps', type=float, default=0.01,
    help='the distance to the origin of the scatter (default: 0.01)')

opts = parser.parse_args()
opts.description = FMT.format(**opts.__dict__)



def load_cfg():
    from src.dict2obj import Config
    from src.base import Adversary
    from src.utils import gpu, load, set_seed
    from models.components import FC, Wrapper

    cfg = Config()
    set_seed(opts.seed)

    # load the model
    arch, dim_feature = load_model(opts.model)
    fc = FC(
        dim_feature=dim_feature,
        num_classes=get_num_classes(opts.dataset)
    )
    arch = arch()
    model = Wrapper(arch=arch, fc=fc)
    model.adv_train(opts.bn_adv) # bn_adv or bn_clean
    print(f"bn_adv: {opts.bn_adv}")
    device = gpu(model)
    load( # load the state dict
        model=model,
        filename=opts.info_path + "/paras.pt",
        device=device, strict=True
    )

    # load the testset
    testset = load_dataset(
        dataset_type=opts.dataset,
        transform='None',
        train=False
    )
    cfg['testloader'] = load_dataloader(
        dataset=testset,
        batch_size=opts.batch_size,
        train=False,
        show_progress=opts.progress
    )
    normalizer = load_normalizer(opts.dataset)

    # generate the log path
    _, log_path = generate_path(METHOD, opts.dataset,
                        opts.model, opts.description)

    # set the 3 attacks for l_inf, l_2, and l_1
    # set the epsilons
    epsilons_linf = torch.linspace(8/255, 1, opts.epsilon_times).tolist()
    epsilons_l2 = torch.linspace(0.25, 0.25, opts.epsilon_times).tolist()
    epsilons_l1 = torch.linspace(7.84, 7.84, opts.epsilon_times).tolist()
    # define the attacks
    attack_linf, bounds_linf, preprocessing_linf = load_attacks(
        attack_type="pgd-linf", dataset_type=opts.dataset,
        stepsize=opts.stepsize, steps=opts.steps
    )
    attack_l2, bounds_l2, preprocessing_l2 = load_attacks(
        attack_type="pgd-l2", dataset_type=opts.dataset,
        stepsize=opts.stepsize, steps=opts.steps
    )
    attack_l1, bounds_l1, preprocessing_l1 = load_attacks(
        attack_type="pgd-l1", dataset_type=opts.dataset,
        stepsize=opts.stepsize, steps=opts.steps
    )
    # define the cfg and actual attack
    cfg['attacker_linf'] = Adversary(
        model=model, attacker=attack_linf, device=device,
        bounds=bounds_linf, preprocessing=preprocessing_linf, epsilon=epsilons_linf
    )
    cfg['attacker_l2'] = Adversary(
        model=model, attacker=attack_l2, device=device,
        bounds=bounds_l2, preprocessing=preprocessing_l2, epsilon=epsilons_l2
    )
    cfg['attacker_l1'] = Adversary(
        model=model, attacker=attack_l1, device=device,
        bounds=bounds_l1, preprocessing=preprocessing_l1, epsilon=epsilons_l1
    )

    cfg['model'] = model
    cfg['normalizer'] = normalizer

    return cfg, log_path

def main(attacker_linf, attacker_l2, attacker_l1, testloader, model, normalizer):
    from src.utils import distance_lp
    datasize = len(testloader.dataset)

    running_success_linf = [0.] * opts.epsilon_times
    running_success_l2 = [0.] * opts.epsilon_times
    running_success_l1 = [0.] * opts.epsilon_times

    running_orig_as_orig = [0.] * opts.epsilon_times

    for i, (inputs, labels) in enumerate(testloader):  # i starts at 0
        inputs = inputs.to(attacker_linf.device)
        labels = labels.to(attacker_linf.device)

        # calculate the clean prediction and accuracy
        outs_clean = model(normalizer(inputs))
        orig_as_orig_count = (outs_clean.argmax(-1) == labels).sum().item()
        for epsilon in range(opts.epsilon_times):
            running_orig_as_orig[epsilon] += orig_as_orig_count

        # create the different adversarial inputs
        # l_inf
        _, clipped_linf, is_adv_linf = attacker_linf(inputs, labels)
        # l_2
        _, clipped_l2, is_adv_l2 = attacker_l2(inputs, labels)
        # l_1
        _, clipped_l1, is_adv_l1 = attacker_l1(inputs, labels)

        dim_ = list(range(1, inputs.dim()))
        for epsilon in range(opts.epsilon_times):
            # select the corresponding pairs, where the respective attack was succesful
            # l_inf
            inputs_linf_ = inputs[is_adv_linf[epsilon]]
            clipped_linf_ = clipped_linf[epsilon][is_adv_linf[epsilon]]
            # l_2
            inputs_l2_ = inputs[is_adv_l2[epsilon]]
            clipped_l2_ = clipped_l2[epsilon][is_adv_l2[epsilon]]
            # l_1
            inputs_l1_ = inputs[is_adv_l1[epsilon]]
            clipped_l1_ = clipped_l1[epsilon][is_adv_l1[epsilon]]

            # track the success of the respective attacks
            # l_inf
            running_success_linf[epsilon] += is_adv_linf[epsilon].sum().item()
            # l_2
            running_success_l2[epsilon] += is_adv_l2[epsilon].sum().item()
            # l_1
            running_success_l1[epsilon] += is_adv_l1[epsilon].sum().item()

            # track the necessary distances
            if i == 0:
                # l_inf distance for samples based on the l_inf attack
                linf_from_linf_list = distance_lp(inputs_linf_, clipped_linf_, p=float('inf'), dim=dim_).cpu().detach().numpy()
                # l_2 distance for samples based on the l_2 attack
                l2_from_l2_list = distance_lp(inputs_l2_, clipped_l2_, p=2, dim=dim_).cpu().detach().numpy()
                # l_inf distance for samples based on the l_2 attack
                linf_from_l2_list = distance_lp(inputs_l2_, clipped_l2_, p=float('inf'), dim=dim_).cpu().detach().numpy()
                # l_1 distance for samples based on the l_1 attack
                l1_from_l1_list = distance_lp(inputs_l1_, clipped_l1_, p=1, dim=dim_).cpu().detach().numpy()
                # l_inf distance for samples based on the l_1 attack
                linf_from_l1_list = distance_lp(inputs_l1_, clipped_l1_, p=float('inf'), dim=dim_).cpu().detach().numpy()

                # plot the manipulation between the original and adversarial input
                import matplotlib.pyplot as plt
                import seaborn as sns
                import numpy as np
                for j in range(10):
                    # for l_inf
                    orig_to_linf_pixel_distances = inputs_linf_[j, :, :, :] - clipped_linf_[j, :, :, :]
                    orig_to_linf_r = orig_to_linf_pixel_distances[0, :, :].cpu().detach().numpy()
                    orig_to_linf_g = orig_to_linf_pixel_distances[1, :, :].cpu().detach().numpy()
                    orig_to_linf_b = orig_to_linf_pixel_distances[2, :, :].cpu().detach().numpy()

                    ax = sns.heatmap(orig_to_linf_r, linewidth = 0.5, cmap = 'coolwarm')
                    plt.savefig('orig_to_linf_r_{}.pdf'.format(j))
                    plt.close()
                    # plt.show()
                    ax = sns.heatmap(orig_to_linf_g, linewidth = 0.5, cmap = 'coolwarm')
                    plt.savefig('orig_to_linf_g_{}.pdf'.format(j))
                    plt.close()
                    # plt.show()
                    ax = sns.heatmap(orig_to_linf_b, linewidth = 0.5, cmap = 'coolwarm')
                    plt.savefig('orig_to_linf_b_{}.pdf'.format(j))
                    plt.close()
                    # plt.show()

                    # for l_2
                    orig_to_l2_pixel_distances = inputs_l2_[j, :, :, :] - clipped_l2_[j, :, :, :]
                    orig_to_l2_r = orig_to_l2_pixel_distances[0, :, :].cpu().detach().numpy()
                    orig_to_l2_g = orig_to_l2_pixel_distances[1, :, :].cpu().detach().numpy()
                    orig_to_l2_b = orig_to_l2_pixel_distances[2, :, :].cpu().detach().numpy()

                    ax = sns.heatmap(orig_to_l2_r, linewidth = 0.5, cmap = 'coolwarm')
                    plt.savefig('orig_to_l2_r_{}.pdf'.format(j))
                    plt.close()
                    # plt.show()
                    ax = sns.heatmap(orig_to_l2_g, linewidth = 0.5, cmap = 'coolwarm')
                    plt.savefig('orig_to_l2_g_{}.pdf'.format(j))
                    plt.close()
                    # plt.show()
                    ax = sns.heatmap(orig_to_l2_b, linewidth = 0.5, cmap = 'coolwarm')
                    plt.savefig('orig_to_l2_b_{}.pdf'.format(j))
                    plt.close()
                    # plt.show()

                    # for l_2
                    orig_to_l1_pixel_distances = inputs_l1_[j, :, :, :] - clipped_l1_[j, :, :, :]
                    orig_to_l1_r = orig_to_l1_pixel_distances[0, :, :].cpu().detach().numpy()
                    orig_to_l1_g = orig_to_l1_pixel_distances[1, :, :].cpu().detach().numpy()
                    orig_to_l1_b = orig_to_l1_pixel_distances[2, :, :].cpu().detach().numpy()

                    ax = sns.heatmap(orig_to_l1_r, linewidth = 0.5, cmap = 'coolwarm')
                    plt.savefig('orig_to_l1_r_{}.pdf'.format(j))
                    plt.close()
                    # plt.show()
                    ax = sns.heatmap(orig_to_l1_g, linewidth = 0.5, cmap = 'coolwarm')
                    plt.savefig('orig_to_l1_g_{}.pdf'.format(j))
                    plt.close()
                    # plt.show()
                    ax = sns.heatmap(orig_to_l1_b, linewidth = 0.5, cmap = 'coolwarm')
                    plt.savefig('orig_to_l1_b_{}.pdf'.format(j))
                    plt.close()
                    # plt.show()

            else:
                # l_inf distance for samples based on the l_inf attack
                linf_from_linf_list = np.append(
                    linf_from_linf_list,
                    distance_lp(inputs_linf_, clipped_linf_, p=float('inf'), dim=dim_).cpu().detach().numpy(),
                    axis=0)
                # l_2 distance for samples based on the l_2 attack
                l2_from_l2_list = np.append(
                    l2_from_l2_list,
                    distance_lp(inputs_l2_, clipped_l2_, p=2, dim=dim_).cpu().detach().numpy(),
                    axis=0)
                # l_inf distance for samples based on the l_2 attack
                linf_from_l2_list = np.append(
                    linf_from_l2_list,
                    distance_lp(inputs_l2_, clipped_l2_, p=float('inf'), dim=dim_).cpu().detach().numpy(),
                    axis=0)
                # l_1 distance for samples based on the l_1 attack
                l1_from_l1_list = np.append(
                    l1_from_l1_list,
                    distance_lp(inputs_l1_, clipped_l1_, p=1, dim=dim_).cpu().detach().numpy(),
                    axis=0)
                # l_inf distance for samples based on the l_1 attack
                linf_from_l1_list = np.append(
                    linf_from_l1_list,
                    distance_lp(inputs_l1_, clipped_l1_, p=float('inf'), dim=dim_).cpu().detach().numpy(),
                    axis=0)

    # plot the histogram of the l_inf distributions
    import matplotlib.pyplot as plt
    import seaborn as sns
    import pandas as pd

    df_linf_from_linf = pd.DataFrame({"linf_from_linf": linf_from_linf_list})
    df_linf_from_l2 = pd.DataFrame({"linf_from_l2": linf_from_l2_list})
    df_linf_from_l1 = pd.DataFrame({"linf_from_l1": linf_from_l1_list})

    fig, ax = plt.subplots(figsize=(12,5))

    df_linf_from_l2.plot.hist(ax=ax, alpha=0.5)
    df_linf_from_l1.plot.hist(ax=ax, alpha=0.5)
    df_linf_from_linf.plot.hist(ax=ax, width=0.005)

    plt.axvline(8/255, 0, 100, c='gray', ls='--')

    plt.savefig('distribution_of_linf_distances.pdf')
    plt.show()
    exit()

    for epsilon in range(opts.epsilon_times):
        running_distance_linf[epsilon] /= running_success[epsilon]
        running_distance_l2[epsilon] /= running_success[epsilon]
        running_success[epsilon] /= datasize

        running_orig_as_orig[epsilon] /= datasize


    running_accuracy = list(map(lambda x: 1. - x, running_success))

    print("Accuracy: ", running_accuracy)
    print("Distance-Linf: ", running_distance_linf)
    print("Distance-L2: ", running_distance_l2)

    print("Clean Accuracy: ", running_orig_as_orig)

    if opts.scatter:
        print("Scatter_eps: ", opts.scatter_eps)
        print("Scatter as adv: ", running_scatter_as_adv)
        print("Scatter as orig: ", running_scatter_as_orig)
        print("Scatter as something: ", running_scatter_as_something)

        print("Adv as orig: ", running_adv_as_orig)
        print("Adv as something: ", running_adv_as_something)


if __name__ == "__main__":
    from torch.utils.tensorboard import SummaryWriter
    from src.utils import mkdirs, readme
    cfg, log_path = load_cfg()
    mkdirs(log_path)
    readme(log_path, opts, mode="a")
    writter = SummaryWriter(log_dir=log_path, filename_suffix=METHOD)

    main(**cfg)

    writter.close()
