


import torch
import torch.nn as nn
import foolbox as fb
import eagerpy as ep
from .utils import AverageMeter, ProgressMeter
from .criteria import DoubleFalse, LogitsAllFalse
from .loss_zoo import cross_entropy, kl_divergence
from models.builder import MoCo

# For the scatter operation
import torch.nn.functional as F
from .config import SCATTER_LEARNING_POLICY
import math

# to calculate the additional time introduced by the scatter operation
import time
import csv

def enter_attack_exit(func):
    def wrapper(attacker, *args, **kwargs):
        attacker.model.attack(True)
        results = func(attacker, *args, **kwargs)
        attacker.model.attack(False)
        return results
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    return wrapper


class Coach:
    """
    Coach is used to train models.
    model: ...
    device: ...
    moco_type: CCC|CAC|CAA|ACC|ACA|AAC|AAA
    leverage: the loss weight
    normalizer: (x - mean) / std
    optimizer: sgd, adam, ...
    learning_policy: learning rate schedule

    ##### New parameters for the scatter operation #####
    scatter: True, to activate the scatter operation
    scatter_k: Number of scattered samples to create
    scatter_eps: Percentage of the distance between original and adversarial input for the scatter operation
    scatter_w: Weight of the scatter loss
    """
    def __init__(
        self, model: MoCo, device,
        moco_type: str, leverage: float,
        normalizer, optimizer, learning_policy,
        ##### Add the new parameters for the scatter operation #####
        scatter=False, scatter_k=10, scatter_eps=0.1, scatter_w=0.5, scatter_lp = 'cosine',
        scatter_steps_stepsize=0.01
    ):
        self.model = model
        self.device = device
        self.normalizer = normalizer
        self.optimizer = optimizer
        self.learning_policy = learning_policy

        self.leverage = leverage
        self.engine = getattr(self, moco_type.lower())

        self.loss = AverageMeter("Loss")
        self.progress = ProgressMeter(self.loss)

        ##### New parameters for the scatter operation #####
        self.scatter = scatter
        self.scatter_k = scatter_k
        self.scatter_eps = scatter_eps
        self.scatter_eta_max = scatter_eps  # for cosine scheduling
        self.scatter_w = scatter_w
        self.scatter_lp = scatter_lp

        self.scatter_steps_stepsize = scatter_steps_stepsize

    def save(self, path):
        torch.save(self.model.encoder_q.state_dict(), path + "/paras.pt")

    def _ccc(self, inputs_q, inputs_k):
        labels = torch.zeros(inputs_q.size(0), dtype=torch.long).to(self.device)
        # c-c-c
        self.model.train()
        self.model.adv_train(False)
        outs_q = self.model.forward_q(self.normalizer(inputs_q))
        outs_k = self.model.forward_k(self.normalizer(inputs_k))
        logits_pos = (outs_q * outs_k).sum(dim=-1, keepdim=True)
        logits_neg = outs_q @ self.model.qc
        logits = torch.cat((logits_pos, logits_neg), dim=1)
        loss_c = cross_entropy(logits, labels, reduction="mean")
        return loss_c, outs_q, outs_k

    def acc(self, inputs_q, inputs_k, attacker):
        n = inputs_q.size(0)
        inputs = torch.cat((inputs_q, inputs_k), dim=0)
        labels = torch.zeros(n, dtype=torch.long).to(self.device)
        criterion = DoubleFalse(labels)

        # track the time to calculate the initial adversarial input
        adv_attack_time = time.time()
        _, inputs_adv, _ = attacker(inputs, criterion)
        adv_attack_time = time.time() - adv_attack_time
        with open('amoc_adv_attack_time.csv','a') as file:
            writer = csv.writer(file, delimiter=',')
            writer.writerow([0, adv_attack_time])

        loss_c, outs_q, outs_k = self._ccc(inputs_q, inputs_k)

        # a-c-c
        self.model.adv_train(True)
        inputs_q_a, inputs_k_a = inputs_adv.split(n)
        outs_q_a = self.model.forward_q(self.normalizer(inputs_q_a))
        outs_k_a = outs_k
        logits_pos_a = (outs_q_a * outs_k_a).sum(dim=-1, keepdim=True)
        logits_neg_a = outs_q_a @ self.model.qc
        logits_a = torch.cat((logits_pos_a, logits_neg_a), dim=1)
        loss_a = cross_entropy(logits_a, labels, reduction="mean")

        ##### New scatter operation #####
        if self.scatter:
            # track the time used to prepare the scatter operation
            scatter_prep_time = time.time()
            loss_scatter = 0

            # calculate the difference between the adversarial and original features
            diff = outs_q_a - outs_q

            # track the time used to prepare the scatter operation
            scatter_prep_time = time.time() - scatter_prep_time
            with open('amoc_scatter_prep_time.csv','a') as file:
                writer = csv.writer(file, delimiter=',')
                writer.writerow([0, scatter_prep_time])

            for _ in range(self.scatter_k):
                scatter_creation_time = time.time()

                # Scatter the adversarial features directly and calculate the loss
                scatter_features = outs_q_a.float().clone()

                # create the random perturbation, scale and apply it to the adversarial features
                rand_perturb = torch.randn(size=diff.size()).cuda()
                rand_perturb *= (self.scatter_eps * diff)

                # Apply the random perturbation to the initial adversarial features
                scatter_features = scatter_features + rand_perturb

                # calculate the loss and track the time used to calculate the loss
                scatter_loss_time = time.time()
                logits_pos_scatter = (scatter_features * outs_k_a).sum(dim=-1, keepdim=True)
                logits_neg_scatter = scatter_features @ self.model.qc
                logits_scatter = torch.cat((logits_pos_scatter, logits_neg_scatter), dim=1)

                loss_scatter += cross_entropy(logits_scatter, labels, reduction="mean")

                # write the time to calculate the loss
                scatter_loss_time = time.time() - scatter_loss_time
                with open('amoc_scatter_loss_time.csv','a') as file:
                    writer = csv.writer(file, delimiter=',')
                    writer.writerow([z, scatter_loss_time])

                # write the time to calculate one pseudo adversarial instance
                scatter_creation_time = time.time() - scatter_creation_time
                with open('amoc_scatter_creation_time.csv','a') as file:
                    writer = csv.writer(file, delimiter=',')
                    writer.writerow([0, scatter_creation_time])

            loss_scatter /= self.scatter_k

        else:
            loss_scatter = None

        return loss_c, loss_a, loss_scatter

    def train(self, trainloader, attacker, epoch=8888):
        self.progress.step() # reset the meter
        for inputs_q, inputs_k, _ in trainloader:
            overall_time = time.time()

            inputs_q = inputs_q.to(self.device)
            inputs_k = inputs_k.to(self.device)

            loss_c, loss_a, loss_scatter = self.engine(inputs_q, inputs_k, attacker)
            loss = self.leverage * loss_c + (1 - self.leverage) * loss_a

            ##### Add the new scatter loss #####
            if loss_scatter is not None:
                loss += self.scatter_w * loss_scatter

            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            self.model.momentum_update_key_encoder() # update the key encoder

            self.loss.update(loss.item(), n=inputs_q.size(0), mode="mean")

            overall_time = time.time() - overall_time
            with open('amoc_overall_time.csv','a') as file:
                writer = csv.writer(file, delimiter=',')
                writer.writerow([0, overall_time])

        self.progress.display(epoch=epoch)
        print("self.scatter_eps: ", self.scatter_eps)
        self.learning_policy.step(epoch=epoch) # update learning rate

        # adapt self.scatter_eps regarding the defined decay method
        if self.scatter:
            if 'cosine' in self.scatter_lp:
                eta_min = SCATTER_LEARNING_POLICY[self.scatter_lp][1]['eta_min']
                T_max = SCATTER_LEARNING_POLICY[self.scatter_lp][1]['T_max']
                # cosine annealing decay
                self.scatter_eps = eta_min + 0.5 * (self.scatter_eta_max - eta_min) * (1 + math.cos((epoch / T_max) * math.pi))

        return self.loss.avg



class LinearCoach:

    def __init__(
        self, model, device,
        normalizer, optimizer,
        learning_policy,
        ##### Add the new parameters for the scatter operation #####
        scatter=False, scatter_k=10, scatter_eps=0.01, scatter_w=1., scatter_lp=None,
        scatter_level='logits', scatter_loss_only=False, scatter_steps_size=0.01
    ):
        self.model = model
        self.device = device
        self.normalizer = normalizer
        self.optimizer = optimizer
        self.learning_policy = learning_policy
        self.loss = AverageMeter("Loss")
        self.acc = AverageMeter("Acc.")
        self.progress = ProgressMeter(self.loss, self.acc)

        ##### New parameters for the scatter operation #####
        self.scatter = scatter
        self.scatter_k = scatter_k
        self.scatter_eps = scatter_eps
        self.scatter_w = scatter_w
        self.scatter_eta_max = scatter_eps  # for cosine scheduling
        self.scatter_lp = scatter_lp

    def save(self, path):
        torch.save(self.model.state_dict(), path + "/paras.pt")


    def train(self, trainlaoder, *, epoch=8888, finetune=False, bn_adv=True):
        self.progress.step() # reset the meter
        self.model.train(finetune)
        self.model.adv_train(bn_adv)
        for inputs, labels in trainlaoder:
            inputs = inputs.to(self.device)
            labels = labels.to(self.device)

            outs = self.model(self.normalizer(inputs))
            loss = cross_entropy(outs, labels, reduction="mean")

            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

            acc_count = (outs.argmax(-1) == labels).sum().item()
            self.loss.update(loss.item(), inputs.size(0), mode="mean")
            self.acc.update(acc_count, inputs.size(0), mode="sum")

        self.progress.display(epoch=epoch)
        self.learning_policy.step()

        return self.loss.avg

    def adv_train(self, trainloader, attacker, *, epoch=8888, finetune=False, bn_adv=True):
        self.progress.step()
        self.model.adv_train(bn_adv)
        for inputs, labels in trainloader:
            overall_time = time.time()

            inputs = inputs.to(self.device)
            labels = labels.to(self.device)

            # track the time to create the initial adversarial input
            adv_attack_time = time.time()
            _, inputs_adv, _ = attacker(inputs, labels)
            adv_attack_time = time.time() - adv_attack_time
            with open('AT_adv_attack_time.csv','a') as file:
                writer = csv.writer(file, delimiter=',')
                writer.writerow([0, adv_attack_time])

            self.model.train(finetune)
            outs, features_adv = self.model(self.normalizer(inputs_adv))

            loss = cross_entropy(outs, labels, reduction="mean")

            ##### New scatter operation #####
            if self.scatter:
                scatter_prep_time = time.time()
                loss_scatter = 0

                # extract the features before the classification head (logits and feature scatter)
                clean_outs, features_orig = self.model(self.normalizer(inputs))

                diff = outs - clean_outs

                # write the time spend on preparing the scatter operation
                scatter_prep_time = time.time() - scatter_prep_time
                with open('AT_scatter_prep_time.csv','a') as file:
                    writer = csv.writer(file, delimiter=',')
                    writer.writerow([0, scatter_prep_time])

                for _ in range(self.scatter_k):
                    # track the time to create one pseudo adversarial instance
                    scatter_creation_time = time.time()

                    scatter_features = outs.float().clone()
                    # random perturbation
                    rand_perturb = torch.randn(size=diff.size()).cuda()
                    rand_perturb *= (self.scatter_eps * diff)

                    # Apply the random perturbation to the initial adversarial features
                    scatter_features = scatter_features + rand_perturb

                    # track the time to calculate the loss
                    scatter_loss_time = time.time()
                    loss_scatter += cross_entropy(scatter_features, labels, reduction="mean")
                    scatter_loss_time = time.time() - scatter_loss_time
                    with open('AT_scatter_loss_time.csv','a') as file:
                        writer = csv.writer(file, delimiter=',')
                        writer.writerow([0, scatter_loss_time])

                    # write the time spend on creating one pseudo adversarial instance
                    scatter_creation_time = time.time() - scatter_creation_time
                    with open('AT_scatter_creation_time.csv','a') as file:
                        writer = csv.writer(file, delimiter=',')
                        writer.writerow([0, scatter_creation_time])

                loss_scatter /= self.scatter_k
                loss += (self.scatter_w * loss_scatter)

            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

            acc_count = (outs.argmax(-1) == labels).sum().item()
            self.loss.update(loss.item(), inputs.size(0), mode="mean")
            self.acc.update(acc_count, inputs.size(0), mode="sum")

            overall_time = time.time() - overall_time
            with open('AT_overall_time.csv','a') as file:
                writer = csv.writer(file, delimiter=',')
                writer.writerow([0, overall_time])

        self.progress.display(epoch=epoch)
        print("self.scatter_eps: ", self.scatter_eps)
        self.learning_policy.step()

        # adapt self.scatter_eps regarding the defined decay method
        if 'cosine' in self.scatter_lp:
            eta_min = SCATTER_LEARNING_POLICY[self.scatter_lp][1]['eta_min']
            T_max = SCATTER_LEARNING_POLICY[self.scatter_lp][1]['T_max']
            # cosine annealing decay
            self.scatter_eps = eta_min + 0.5 * (self.scatter_eta_max - eta_min) * (1 + math.cos((epoch / T_max) * math.pi))

        return self.loss.avg

class Adversary:
    """
    Adversary is mainly based on foolbox, especially pytorchmodel.
    model: Make sure that the model's output is the logits or the attack is adapted.
    attacker: the attack implemented by foolbox or a similar one
    device: ...
    bounds: typically [0, 1]
    preprocessing: including mean, std, which is similar to normalizer
    criterion: typically given the labels and consequently it is Misclassification,
            other critera could be given to carry target attack or black attack.
    """
    def __init__(
        self, model, attacker, device,
        bounds, preprocessing, epsilon
    ):
        model.eval()
        self.fmodel = fb.PyTorchModel(
            model,
            bounds=bounds,
            preprocessing=preprocessing,
            device=device
        )
        self.model = model
        self.device = device
        self.epsilon = epsilon
        self.attacker = attacker

    def attack(self, inputs, criterion, epsilon=None):
        if epsilon is None:
            epsilon = self.epsilon
        self.model.eval() # make sure in evaluating mode ...
        return self.attacker(self.fmodel, inputs, criterion, epsilons=epsilon)

    @torch.no_grad()
    def accuracy(self, inputs, labels):
        inputs_, labels_ = ep.astensors(inputs, labels)
        del inputs, labels

        self.model.eval() # make sure in evaluating mode ...
        predictions = self.fmodel(inputs_).argmax(axis=-1)
        accuracy = (predictions == labels_)
        return accuracy.sum().item()

    def success(self, inputs, criterion, epsilon=None):
        _, _, is_adv = self.attack(inputs, criterion, epsilon)
        return is_adv.sum().item()

    def evaluate(self, dataloader, epsilon=None, bn_adv=True):
        datasize = len(dataloader.dataset)
        running_accuracy = 0
        running_success = 0
        self.model.adv_train(bn_adv)
        for inputs, labels in dataloader:
            inputs = inputs.to(self.device)
            labels = labels.to(self.device)
            running_accuracy += self.accuracy(inputs, labels)
            running_success += self.success(inputs, labels, epsilon)
        return running_accuracy / datasize, running_success / datasize

    @enter_attack_exit
    def __call__(self, inputs, criterion, *, epsilon=None):
        return self.attack(inputs, criterion, epsilon)



class FBDefense:
    def __init__(self, model, device, bounds, preprocessing):
        self.rmodel = fb.PyTorchModel(
            model,
            bounds=bounds,
            preprocessing=preprocessing,
            device=device
        )

        self.model = model

    @torch.no_grad()
    def query(self, inputs):
        self.model.eval()
        return self.rmodel(inputs)

    def __call__(self, inputs):
        return self.query(inputs)
