
import torch
import numpy as np
import os
from PIL import Image
from .config import ROOT


# For Cifar10 additional data usage
import pickle


# https://github.com/VITA-Group/Adversarial-Contrastive-Learning/blob/937019219497b449f4cb61cc6118fdf32cc3de12/data/cifar10_c.py#L9
class CIFAR10C(torch.utils.data.Dataset):
    filename = "CIFAR-10-C"
    def __init__(self, root=ROOT, transform=None, severity=5, corruption_type=''):
        root = os.path.join(root, self.filename)
        dataPath = os.path.join(root, '{}.npy'.format(corruption_type))
        labelPath = os.path.join(root, 'labels.npy')

        self.data = np.load(dataPath)
        self.label = np.load(labelPath).astype(np.long)
        self.transform = transform

    def __getitem__(self, idx):

        img = self.data[idx]
        img = Image.fromarray(img).convert('RGB')

        if self.transform is not None:
            img = self.transform(img)

        label = self.label[idx]

        return img, label

    def __len__(self):
        return self.data.shape[0]


class CIFAR100C(CIFAR10C):
    filename = "CIFAR-100-C"


# Load Cifar-10 and additional data, as proposed by https://github.com/deepmind/deepmind-research/tree/master/adversarial_robustness
class CIFAR10_Plus(torch.utils.data.Dataset):

    base_folder = 'cifar-10-batches-py'
    url = "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
    filename = "cifar-10-python.tar.gz"
    tgz_md5 = 'c58f30108f718f92721af3b95e74349a'
    train_list = [
        ['data_batch_1', 'c99cafc152244af753f735de768cd75f'],
        ['data_batch_2', 'd4bba439e000b95fd0a9bffe97cbabec'],
        ['data_batch_3', '54ebc095f3ab1f0389bbae665268c751'],
        ['data_batch_4', '634d18415352ddfa80567beed471001a'],
        ['data_batch_5', '482c414d41f54cd18b22e5b47cb7c3cb'],
    ]

    test_list = [
        ['test_batch', '40351d587109b95175f43aff81a1287e'],
    ]
    meta = {
        'filename': 'batches.meta',
        'key': 'label_names',
        'md5': '5ff9c542aee3614f3951f8cda6e48888',
    }

    filename = "CIFAR10_Plus"

    def __init__(self, root=ROOT, train=True, transform=None, download=True):

        ##### Load the original cifar10 data #####
        self.root = root
        self.train = train  # training set or test set

        if self.train:
            downloaded_list = self.train_list
        else:
            downloaded_list = self.test_list

        self.data: Any = []
        self.label = []

        # now load the picked numpy arrays
        for file_name, checksum in downloaded_list:
            file_path = os.path.join(self.root, self.base_folder, file_name)
            with open(file_path, 'rb') as f:
                entry = pickle.load(f, encoding='latin1')
                self.data.append(entry['data'])
                if 'labels' in entry:
                    self.label.extend(entry['labels'])
                else:
                    self.label.extend(entry['fine_labels'])

        self.data = np.vstack(self.data).reshape(-1, 3, 32, 32)
        self.data = self.data.transpose((0, 2, 3, 1))  # convert to HWC

        # Convert label from list to np.array
        self.label = np.array(self.label)

        ##### Load the additional data according to https://github.com/deepmind/deepmind-research/tree/master/adversarial_robustness #####
        add_dataPath = os.path.join(root, 'cifar10_ddpm/image.npy')
        add_labelPath = os.path.join(root, 'cifar10_ddpm/label.npy')

        # load the additional data and labels
        add_data = np.load(add_dataPath)
        add_label = np.load(add_labelPath)

        # stack the original and additional data and label
        self.data = np.vstack((self.data, add_data))
        self.label = np.concatenate((self.label, add_label))

        self.transform = transform

    def __getitem__(self, idx):

        img, label = self.data[idx], self.label[idx]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        return img, label

    def __len__(self):
        return self.data.shape[0]



class SingleSet(torch.utils.data.Dataset):

    def __init__(self, dataset, transform):
        self.data = dataset
        self.transform = transform

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        img, label = self.data[index]
        return self.transform(img), label

class DoubleSet(SingleSet):

    def __getitem__(self, index):
        img, label = self.data[index]
        img1 = self.transform(img)
        img2 = self.transform(img)
        return img1, img2, label
