#!/usr/bin/env python3 -u

from __future__ import print_function

import csv
import os
import math
import time

import torch
import torch.backends.cudnn as cudnn
import torch.nn as nn

import torch.optim as optim
import data_loader
import model_loader

from attack_lib import FastGradientSignUntargeted,RepresentationAdv

from models.projector import Projector
from argument import parser, print_args
from utils import progress_bar, checkpoint, AverageMeter, accuracy

from loss import pairwise_similarity, NT_xent
from torchlars import LARS
from warmup_scheduler import GradualWarmupScheduler

import config

torch.autograd.set_detect_anomaly(True)

args = parser()

def print_status(string):
    if args.local_rank % ngpus_per_node == 0:
    # if os.environ('LOCAL_RANK') % ngpus_per_node == 0:
        print(string)

ngpus_per_node = torch.cuda.device_count()
if args.ngpu>1:
    multi_gpu=True
elif args.ngpu==1:
    multi_gpu=False
else:
    assert("Need GPU....")
if args.local_rank % ngpus_per_node == 0:
    print_args(args)

start_epoch = 0  # start from epoch 0 or last checkpoint epoch

if args.seed != 0:
    torch.manual_seed(args.seed)

world_size = args.ngpu
torch.distributed.init_process_group(
    'nccl',
    init_method='env://',
    world_size=world_size,
    rank=args.local_rank,
    # rank=os.environ('LOCAL_RANK'),
)

# Data
print_status('==> Preparing data..')
if not (args.train_type=='contrastive'):
    assert('wrong train phase...')
else:
    trainloader, traindst, testloader, testdst ,train_sampler = data_loader.get_dataset(args)

# Model
print_status('==> Building model..')
torch.cuda.set_device(args.local_rank)
# torch.cuda.set_device(os.environ('LOCAL_RANK'))
model = model_loader.get_model(args)

if args.model=='ResNet18':
    expansion=1
elif args.model=='ResNet50':
    expansion=4
else:
    assert('wrong model type')
projector = Projector(expansion=expansion)

if 'Rep' in args.advtrain_type:
    # Rep_info = 'Rep_attack_ep_'+str(args.epsilon)+'_alpha_'+ str(args.alpha) + '_min_val_' + str(args.min) + '_max_val_' + str(args.max) + '_max_iters_' + str(args.k) + '_type_' + str(args.attack_type) + '_randomstart_' + str(args.random_start)
    Rep_info = 'ins_'
    args.name += Rep_info

    print_status("Representation attack info ...")
    print_status(Rep_info)

    Rep = RepresentationAdv(model=model,
                            projector=projector,
                            epsilon=args.epsilon,
                            alpha=args.alpha,
                            min_val=args.min,
                            max_val=args.max,
                            max_iters=args.k,
                            _type=args.attack_type,
                            loss_type=args.loss_type,
                            regularize = args.regularize_to)
else:
    assert('wrong adversarial train type')

if args.scatter:
    Rep_info = 'scatter_'
    args.name += Rep_info

# Model upload to GPU #
model.cuda()
projector.cuda()

model       = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model)
model       = torch.nn.parallel.DistributedDataParallel(
                model,
                device_ids=[args.local_rank],
                output_device=args.local_rank,
                find_unused_parameters=True,
)
projector   = torch.nn.parallel.DistributedDataParallel(
                projector,
                device_ids=[args.local_rank],
                output_device=args.local_rank,
                find_unused_parameters=True,
)

cudnn.benchmark = True
print_status('Using CUDA..')

# Aggregating model parameter & projection parameter #
model_params = []
model_params += model.parameters()
model_params += projector.parameters()

# LARS optimizer from KAKAO-BRAIN github "pip install torchlars" or git from https://github.com/kakaobrain/torchlars
base_optimizer  = optim.SGD(model_params, lr=args.lr, momentum=0.9, weight_decay=args.decay)
optimizer       = LARS(optimizer=base_optimizer, eps=1e-8, trust_coef=0.001)

# Cosine learning rate annealing (SGDR) & Learning rate warmup git from https://github.com/ildoonet/pytorch-gradual-warmup-lr #
scheduler_cosine = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, args.epoch)
scheduler_warmup = GradualWarmupScheduler(optimizer, multiplier=args.lr_multiplier, total_epoch=10, after_scheduler=scheduler_cosine)

# if scatter cosine decay, set scatter_eta_max at the starting scatter eps
SCATTER_ETA_MAX = args.scatter_eps

def train(epoch):
    if args.local_rank % ngpus_per_node == 0:
        print('\nEpoch: %d' % epoch)

    model.train()
    projector.train()

    train_sampler.set_epoch(epoch)

    # adjust the learning rates accorrdingly
    scheduler_warmup.step()

    total_loss = 0
    reg_simloss = 0
    reg_advloss = 0
    reg_scatterloss = 0

    for batch_idx, (ori, inputs_1, inputs_2, label) in enumerate(trainloader):
        overall_time = time.perf_counter()

        ori, inputs_1, inputs_2 = ori.cuda(), inputs_1.cuda() ,inputs_2.cuda()

        # Execute the adversarial attack
        if args.attack_to=='original':
            attack_target = inputs_1
        else:
            attack_target = inputs_2

        adv_attack_time = time.perf_counter()
        if 'Rep' in args.advtrain_type :
            advinputs, adv_loss = Rep.get_loss(
                original_images=inputs_1,
                target = attack_target,
                optimizer=optimizer,
                weight= args.lamda,
                random_start=args.random_start)
            reg_advloss    += adv_loss.data
        adv_attack_time = time.perf_counter() - adv_attack_time
        with open('RoCL_adv_attack_time.csv','a') as file:
            writer = csv.writer(file, delimiter=',')
            writer.writerow([0, adv_attack_time])

        if not (args.advtrain_type == 'None'):
            inputs = torch.cat((inputs_1, inputs_2, advinputs))
        else:
            inputs = torch.cat((inputs_1, inputs_2))

        outputs = projector(model(inputs))

        # Calculate the SimCLR loss for the clean, and adversarial inputs
        similarity, gathered_outputs = pairwise_similarity(outputs,
            temperature=args.temperature, multi_gpu=multi_gpu,
            adv_type = args.advtrain_type)

        simloss  = NT_xent(similarity, args.advtrain_type)
        reg_simloss += simloss.data

        # update the overall loss
        loss = simloss

        # add the adversarial loss to the overall loss
        if not (args.advtrain_type=='None'):
            # if not (args.scatter_loss_only):
            loss += adv_loss

        # perform the scatter operation on the adversarial inputs
        if args.scatter:
            scatter_prep_time = time.time()
            scatterloss = 0

            # extract the features of the clean and adversarial inputs
            advfeatures = outputs[2 * inputs_1.size(0):]
            origfeatures = outputs[:2 * inputs_1.size(0)]

            # split the original features, to calculate the difference regarding
            # the attacked input features
            orig_1_features = origfeatures[:inputs_1.size(0)]
            orig_2_features = origfeatures[inputs_1.size(0):]

            # calculate the difference between the adversarial and its reference features
            # TODO: if the current results are not good, try to use orig_1_features always.
            # Actually, this would make more sense, as the adversarial input is always created based on inputs_1
            # if args.attack_to=='original':
            diff = advfeatures - orig_1_features
            # else:
            #     diff = advfeatures - orig_2_features

            scatter_prep_time = time.time() - scatter_prep_time
            with open('RoCL_scatter_prep_time.csv','a') as file:
                writer = csv.writer(file, delimiter=',')
                writer.writerow([0, scatter_prep_time])

            # create scatter_k different scattered versions of the adversarial input
            for _ in range(args.scatter_k):
                scatter_creation_time = time.time()
                # Scatter the adversarial features directly and calculate the loss
                scatter_features = advfeatures.float().clone()

                # random perturbation
                # rand_perturb = torch.FloatTensor(scatter_features.shape).uniform_(
                #     -current_eps, current_eps).float().cuda()

                # create the random perturbation, scale and apply it to the adversarial features
                rand_perturb = torch.randn(size=diff.size()).cuda()
                rand_perturb *= (args.scatter_eps * diff)
                scatter_features += rand_perturb

                # # Apply the random perturbation to the initial adversarial features
                # scatter_features = scatter_features + rand_perturb

                # normalize the scatter features
                norm = scatter_features.pow(2).sum(1, keepdim=True).pow(1./2)
                scatter_features = scatter_features.div(norm)

                # combine regular and scattered features and calculate the loss
                outputs = torch.cat((origfeatures, scatter_features))

                scatter_loss_time = time.time()
                similarity, gathered_outputs = pairwise_similarity(outputs,
                    temperature=args.temperature, multi_gpu=multi_gpu,
                    adv_type = 'Rep')

                scatterloss += NT_xent(similarity, 'Rep')
                scatter_loss_time = time.time() - scatter_loss_time
                with open('RoCL_scatter_loss_time.csv','a') as file:
                    writer = csv.writer(file, delimiter=',')
                    writer.writerow([0, scatter_loss_time])

                scatter_creation_time = time.time() - scatter_creation_time
                with open('RoCL_scatter_creation_time.csv','a') as file:
                    writer = csv.writer(file, delimiter=',')
                    writer.writerow([0, scatter_creation_time])

            # for scatter + adv loss
            # scatterloss += adv_loss
            scatterloss /= (args.scatter_k)
            scatterloss *= args.scatter_w
            reg_scatterloss += scatterloss.data

            # add to the total loss
            loss += scatterloss

        # Update total loss for display
        total_loss += loss.data

        # Optimization step
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if (args.local_rank % ngpus_per_node == 0):
            # Build the string to plot
            string = "Total: {:.2f}".format(total_loss / (batch_idx + 1))
            string += " | Sim: {:.3f}".format(reg_simloss / (batch_idx + 1))
            string += " | lr: {}".format(optimizer.param_groups[0]['lr'])
            if 'Rep' in args.advtrain_type:
                string += " | Adv: {:.3f}".format(reg_advloss / (batch_idx + 1))
            if args.scatter:
                string += " | Scatter: {:.3f}".format(reg_scatterloss / (batch_idx + 1))

            progress_bar(batch_idx, len(trainloader), string)

        overall_time = time.perf_counter() - overall_time
        with open('RoCL_overall_time.csv','a') as file:
            writer = csv.writer(file, delimiter=',')
            writer.writerow([0, overall_time])

    return (total_loss/batch_idx, reg_simloss/batch_idx)


def test(epoch, train_loss):
    model.eval()
    projector.eval()

    # Save at the last epoch #
    if epoch == args.epoch - 1 and args.local_rank % ngpus_per_node == 0:
        checkpoint(model, train_loss, epoch, args, optimizer)
        checkpoint(projector, train_loss, epoch, args, optimizer, save_name_add='_projector')

    # Save at every 100 epoch #
    elif epoch % 100 == 0 and args.local_rank % ngpus_per_node == 0:
            checkpoint(model, train_loss, epoch, args, optimizer, save_name_add='_epoch_'+str(epoch))
            checkpoint(projector, train_loss, epoch, args, optimizer, save_name_add=('_projector_epoch_' + str(epoch)))


# Log and saving checkpoint information #
if not os.path.isdir('results') and args.local_rank % ngpus_per_node == 0:
    os.mkdir('results')

args.name += (args.train_type + '_' +args.model + '_' + args.dataset + '_b' + str(args.batch_size)+'_nGPU'+str(args.ngpu)+'_l'+str(args.lamda))
loginfo = 'results/log_' + args.name + '_' + str(args.seed)
logname = (loginfo+ '.csv')
print_status('Training info...')
print_status(loginfo)

##### Log file #####
if args.local_rank % ngpus_per_node == 0:
    with open(logname, 'w') as logfile:
        logwriter = csv.writer(logfile, delimiter=',')
        logwriter.writerow(['epoch', 'train loss', 'reg loss'])

print(args.name)

##### Training #####
for epoch in range(start_epoch, args.epoch):
    train_loss, reg_loss = train(epoch)
    test(epoch, train_loss)

    # update the scatter_eps value according to the defined decay method
    if args.scatter:
        if 'stepwise' in args.scatter_lp:
            if epoch in config.SCATTER_LEARNING_POLICY[args.scatter_lp][1]['milestones']:
                args.scatter_eps -= config.SCATTER_LEARNING_POLICY[args.scatter_lp][1]['gamma']
        elif 'cosine' in args.scatter_lp:
            eta_min = config.SCATTER_LEARNING_POLICY[args.scatter_lp][1]['eta_min']
            T_max = config.SCATTER_LEARNING_POLICY[args.scatter_lp][1]['T_max']
            # cosine annealing decay
            args.scatter_eps = eta_min + 0.5 * (SCATTER_ETA_MAX - eta_min) * (1 + math.cos((epoch / T_max) * math.pi))

        print("args.scatter_eps: ", args.scatter_eps)

    if args.local_rank % ngpus_per_node == 0:
        with open(logname, 'a') as logfile:
            logwriter = csv.writer(logfile, delimiter=',')
            logwriter.writerow([epoch, train_loss.item(), reg_loss.item()])
