#!/usr/bin/env python

import torch
import argparse
from src.loadopts import *

# to work here
import data_loader
import model_loader
import models
from models.projector import Projector
from collections import OrderedDict
import torch.nn as nn

METHOD = "WhiteBox"
FMT = "{description}={bn_adv}={attack}-{stepsize}-{steps}" \
        "={epsilon_min}-{epsilon_max}-{epsilon_times}"


parser = argparse.ArgumentParser()
parser.add_argument("model", type=str)
parser.add_argument("dataset", type=str)
parser.add_argument("info_path", type=str)
parser.add_argument("--bn_adv", action="store_false", default=True)
parser.add_argument("--attack", type=str, default="pgd-linf")
parser.add_argument("--epsilon_min", type=float, default=8/255)
parser.add_argument("--epsilon_max", type=float, default=1.)
parser.add_argument("--epsilon_times", type=int, default=1)
parser.add_argument("--stepsize", type=float, default=0.1,
                    help="pgd:rel_stepsize, cwl2:step_size, deepfool:overshoot, bb:lr")
parser.add_argument("--steps", type=int, default=20)
parser.add_argument("-b", "--batch_size", type=int, default=128)
parser.add_argument("--progress", action="store_false", default=True,
                help="show the progress if true")
parser.add_argument("--seed", type=int, default=1)
parser.add_argument("-m", "--description", type=str, default="attack")

parser.add_argument('--scatter', action='store_true',
    help='if set, scatter the adversarial sample in feature space')
parser.add_argument('--scatter_eps', type=float, default=0.01,
    help='the distance to the origin of the scatter (default: 0.01)')

# Necessary to work with the RoCL models
parser.add_argument('--train_type', default='linear_eval', type=str, help='standard')
parser.add_argument('--color_jitter_strength', default=0.5, type=float, help='0.5 for CIFAR, 1.0 for ImageNet')

opts = parser.parse_args()
opts.description = FMT.format(**opts.__dict__)


# output configs for the backbone and classifier
if opts.dataset == 'cifar10' or opts.dataset=='mnist':
    num_outputs = 10
elif opts.dataset == 'cifar100':
    num_outputs = 100

if opts.model == 'ResNet50':
    expansion = 4
else:
    expansion = 1

def load_cfg():
    from src.dict2obj import Config
    from src.base import Adversary
    from src.utils import gpu, load, set_seed
    from models.components import FC, Wrapper

    cfg = Config()
    set_seed(opts.seed)

    # load the RoCL backbone
    model = model_loader.get_model(opts)

    checkpoint_ = torch.load(opts.info_path)

    new_state_dict = OrderedDict()
    for k, v in checkpoint_['model'].items():
        name = k[7:]
        new_state_dict[name] = v

    model.load_state_dict(new_state_dict)

    # define the linear classifier
    if opts.dataset=='cifar-10':
        Linear = nn.Sequential(nn.Linear(512*expansion, 10))
    elif opts.dataset=='cifar-100':
        Linear = nn.Sequential(nn.Linear(512*expansion, 100))

    linearcheckpoint_ = torch.load(opts.info_path+'_linear')
    new_state_dict = OrderedDict()
    for k, v in linearcheckpoint_['model'].items():
        name = k[7:] # remove `module.`
        new_state_dict[name] = v
    Linear.load_state_dict(new_state_dict)

    # wrap backbone and classifier to one model
    model = Wrapper(arch=model, fc=Linear)

    # model.adv_train(opts.bn_adv) # bn_adv or bn_clean
    model.adv_train(False) # bn_adv or bn_clean
    print(f"bn_adv: {opts.bn_adv}")
    device = gpu(model)

    # Data
    trainloader, traindst, testloader, testdst  = data_loader.get_dataset(opts)
    cfg['testloader'] = load_dataloader(
        dataset=testdst,
        batch_size=opts.batch_size,
        train=False,
        show_progress=opts.progress
    )

    # generate the log path
    _, log_path = generate_path(METHOD, opts.dataset,
                        opts.model, opts.description)

    # set the attack
    attack, bounds, preprocessing = load_attacks(
        attack_type=opts.attack, dataset_type=opts.dataset,
        stepsize=opts.stepsize, steps=opts.steps
    )

    epsilons = torch.linspace(opts.epsilon_min, opts.epsilon_max, opts.epsilon_times).tolist()
    cfg['attacker'] = Adversary(
        model=model, attacker=attack, device=device,
        bounds=bounds, preprocessing=None, epsilon=epsilons
    )

    cfg['model'] = model
    # cfg['normalizer'] = normalizer

    return cfg, log_path

def main(attacker, testloader, model):
    from src.utils import distance_lp
    running_success = [0.] * opts.epsilon_times
    running_distance_linf = [0.] * opts.epsilon_times
    running_distance_l2 = [0.] * opts.epsilon_times
    running_orig_as_orig = [0.] * opts.epsilon_times

    # track the scatter stats
    if opts.scatter:
        running_scatter_as_orig = [0.] * opts.epsilon_times
        running_scatter_as_adv = [0.] * opts.epsilon_times
        running_scatter_as_something = [0.] * opts.epsilon_times

        running_adv_as_orig = [0.] * opts.epsilon_times
        running_adv_as_something = [0.] * opts.epsilon_times

    for inputs, labels in testloader:
        inputs = inputs.to(attacker.device)
        labels = labels.to(attacker.device)

        # calculate the clean prediction and accuracy
        outs_clean = model(inputs)
        orig_as_orig_count = (outs_clean.argmax(-1) == labels).sum().item()
        for epsilon in range(opts.epsilon_times):
            running_orig_as_orig[epsilon] += orig_as_orig_count

        _, clipped, is_adv = attacker(inputs, labels)

        dim_ = list(range(1, inputs.dim()))
        for epsilon in range(opts.epsilon_times):
            inputs_ = inputs[is_adv[epsilon]]
            clipped_ = clipped[epsilon][is_adv[epsilon]]

            running_success[epsilon] += is_adv[epsilon].sum().item()
            running_distance_linf[epsilon] += distance_lp(inputs_, clipped_, p=float('inf'), dim=dim_).sum().item()
            running_distance_l2[epsilon] += distance_lp(inputs_, clipped_, p=2, dim=dim_).sum().item()

        if opts.scatter:
            inputs_adv = clipped[0]

            # calculate the adversarial predictions
            outs_adv = model(inputs_adv)
            adv_as_orig_count = (outs_adv.argmax(-1) == labels).sum().item()
            adv_as_something_count = (outs_adv.argmax(-1) != labels).sum().item()

            for epsilon in range(opts.epsilon_times):
                running_adv_as_orig[epsilon] += adv_as_orig_count
                running_adv_as_something[epsilon] += adv_as_something_count

            # calculate the difference between adversarial and clean outputs
            outs_orig = model(normalizer(inputs))
            diff = outs_adv - outs_orig

            # scatter the adversarial outputs
            scatter_features = outs_adv.float().clone()

            # normal distributed random perturbation
            rand_perturb = torch.randn(size=diff.size()).cuda()

            # Scale and apply the random perturbation to the initial adversarial features
            scatter_features += (rand_perturb * diff * opts.scatter_eps)

            # prediction: scatter as orig
            scatter_as_orig_count = (scatter_features.argmax(-1) == labels).sum().item()
            # prediction: scatter as adv
            scatter_as_adv_count = (scatter_features.argmax(-1) == outs_adv.argmax(-1)).sum().item()
            # prediction: scatter as something else
            scatter_as_something_count = (scatter_features.argmax(-1) != labels).sum().item()

            for epsilon in range(opts.epsilon_times):
                running_scatter_as_orig[epsilon] += scatter_as_orig_count
                running_scatter_as_adv[epsilon] += scatter_as_adv_count
                running_scatter_as_something[epsilon] += scatter_as_something_count


    datasize = len(testloader.dataset)
    head = "-".join(map(str, (opts.attack, opts.epsilon_min, opts.epsilon_max,
                        opts.epsilon_times, opts.stepsize, opts.steps)))
    for epsilon in range(opts.epsilon_times):
        running_distance_linf[epsilon] /= running_success[epsilon]
        running_distance_l2[epsilon] /= running_success[epsilon]
        running_success[epsilon] /= datasize

        running_orig_as_orig[epsilon] /= datasize

        # writter.add_scalar(head+"Success", running_success[epsilon], epsilon)
        # writter.add_scalars(
        #     head+"Distance",
        #     {
        #         "Linf": running_distance_linf[epsilon],
        #         "L2": running_distance_l2[epsilon],
        #     },
        #     epsilon
        # )

        if opts.scatter:
            running_scatter_as_orig[epsilon] /= datasize
            running_scatter_as_adv[epsilon] /= datasize
            running_scatter_as_something[epsilon] /= datasize

            running_adv_as_orig[epsilon] /= datasize
            running_adv_as_something[epsilon] /= datasize

    running_accuracy = list(map(lambda x: 1. - x, running_success))

    print("Accuracy: ", running_accuracy)
    print("Distance-Linf: ", running_distance_linf)
    print("Distance-L2: ", running_distance_l2)

    print("Clean Accuracy: ", running_orig_as_orig)

    if opts.scatter:
        print("Scatter_eps: ", opts.scatter_eps)
        print("Scatter as adv: ", running_scatter_as_adv)
        print("Scatter as orig: ", running_scatter_as_orig)
        print("Scatter as something: ", running_scatter_as_something)

        print("Adv as orig: ", running_adv_as_orig)
        print("Adv as something: ", running_adv_as_something)


if __name__ == "__main__":
    from torch.utils.tensorboard import SummaryWriter
    from src.utils import mkdirs, readme
    cfg, log_path = load_cfg()
    mkdirs(log_path)
    readme(log_path, opts, mode="a")
    writter = SummaryWriter(log_dir=log_path, filename_suffix=METHOD)

    main(**cfg)

    writter.close()
